
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "rhp.h"
#include "sif_f.h"

extern void	Error(char *msg);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error(#F " : multi-threading initialization error.")

#define TERM_MT(T,F)	\
	if (TERM_THREAD(T)) Error(#F " : multi-threading termination error.")

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

#ifndef	Float
#define Float	float
#endif

#ifndef	Filter
#define Filter	Chesler
#endif

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}

extern double	Filter(int j);

typedef struct {
		double	x,y;
	} Vector;

typedef struct {
		int	j;
		double	r;
	} Struct;

static int	N,L,N1,N2,Threads=1;
static Vector	*E,*G,**W;
static double	**P,**Q,f4cbp;
static Struct	**S4BP;
static Float	***F;

static void	FFT(int sign,Vector *V)
{
	int	j,k,l,m,p,q;
	Vector	v,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		v=V[j]; V[j]=V[k]; V[k]=v;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=V+p, q=0; q<l; q++, a+=j) {
		    v=(*(b=a+k)); b->x=a->x-(z=v.x*x-v.y*y); a->x+=z;
				  b->y=a->y-(z=v.x*y+v.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T	Convolution(void *i)
{
	size_t	l;
	Vector	*v,*g,w,*V=W[(size_t)i];
	double	*p;
	int	j;

	for (l=(size_t)i; l<L; l+=Threads) {
	    for (v=V, j=0; j<N1; j++, v++) v->x=v->y=0.0;

	    p=P[l];
	    for (j=0; j<N; j++, v++) {
		v->x=(*(p++)); v->y=0.0;
	    }
	    for (; j<N1; j++, v++) v->x=v->y=0.0;

	    FFT(-1,V);

	    for (v=V, g=G, j=0; j<N2; j++, v++, g++) {
		w=(*v); v->x=g->x*w.x-g->y*w.y;
			v->y=g->x*w.y+g->y*w.x;
	    }
	    FFT(1,V);

	    for (v=V, j=0; j<N; j++, v++) Q[j][l]=f4cbp*v->x;
	}
	return RETURN_VALUE;
}

static FUNCTION_T	BackProjection(void *i)
{
	size_t	y;
	int	x,j,z;
	Float	**Fy,*Fyx;
	Struct	*s4bp;
	double	*Q0,*Q1,r0,r1;

	for (y=(size_t)i; y<N; y+=Threads) {
	    Fy=F[y]; s4bp=S4BP[y];
	    for (x=0; x<N; x++)
		if ((j=s4bp[x].j)>=0) {
		    Fyx=Fy[x]; Q0=Q[j]; Q1=Q[j+1]; r0=1.0-(r1=s4bp[x].r);
		    for (z=0; z<L; z++) *(Fyx++)+=(*(Q0++)*r0+(*(Q1++))*r1);
		}
	}
	return RETURN_VALUE;
}

static double	Log(double d)
{
	return (d>0.0)?log(d):0.0;
}

#define LEN	2048
#define EPS	1e-9

int	main(int argc,char **argv)
{
	HiPic	hp;
	int	Ox,Oy,M;

	if (argc!=4 && argc!=7 && argc!=9)
	Error("usage : oct_tg HiPic/ Ox Oy {{layer1 layer2} Dr RA0 TG_format}");

	InitReadHiPic(argv[1],&hp);

	if (hp.Nt%2) (void)fputs("bad number of views (warning).\n",stderr);

	if ((Ox=atoi(argv[2]))+hp.Nx<=0 || hp.Nx<=Ox ||
	    (Oy=atoi(argv[3]))+hp.Ny<=0 || hp.Ny<=Oy) Error("bad offset.");

	(void)fprintf(stderr,"%d\t%d\t%d\n",N=hp.Nx+abs(Ox),
					    L=hp.Ny-abs(Oy),
					    M=hp.Nt/2);

	if (argc!=4)
{
	char		*env,path[LEN];
	int		z2,z,r,y,x,m;
	double		Dr,ra0,*p,S,C,O,R,
			Or=(double)(N-1)/2.0;
	THREAD_T	*Thread;
	size_t		i;
	FOM		*t,**f;

	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	if (argc==7)
	    z2=0;
	else {
	    if ((z2=atoi(argv[4]))<0 ||
		(z=atoi(argv[5]))>=L || z2>z) Error("bad layer range.");

	    L=z-z2+1;
	}
	if ((Dr=atof(argv[argc-3]))<EPS) Error("bad interval of detectors.");

	for (N1=1; N1<N; N1<<=1) ; N2=N1<<1;

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (W=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*W=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (P=(double **)malloc(sizeof(double *)*L))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*L*N))==NULL ||
	    (Q=(double **)malloc(sizeof(double *)*(N+1)))==NULL ||
	    (*Q=(double *)malloc(sizeof(double)*N*L))==NULL ||
	    (S4BP=(Struct **)malloc(sizeof(Struct *)*N))==NULL ||
	    (*S4BP=(Struct *)malloc(sizeof(Struct)*N*N))==NULL ||
	    (F=(Float ***)malloc(sizeof(Float **)*N))==NULL ||
	    (*F=(Float **)malloc(sizeof(Float *)*N*N))==NULL ||
	    (**F=(Float *)malloc(sizeof(Float)*N*N*L))==NULL)
	    Error("memory allocation error.");

	for (r=0; r<N2; r++) {
	    sincos(M_PI*(double)r/(double)N2,&(E[r].y),&(E[r].x));

	    G[r].x=Filter(r-N1); G[r].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) W[i]=W[i-1]+N2;

	for (z=1; z<L; z++) P[z]=P[z-1]+N;

	Q[N]=Q[0]; for (r=1; r<N; r++) Q[r]=Q[r-1]+L;

	for (y=0; y<N; y++) {
	    if (y>0) {
		S4BP[y]=S4BP[y-1]+N; F[y]=F[y-1]+N;
	    }
	    for (x=0; x<N; x++) {
		S4BP[y][x].j=(hypot((double)x-Or,Or-(double)y)<=Or)-1;

		if (y>0 || x>0) F[y][x]=F[y][x-1]+L;

		for (z=0; z<L; z++) F[y][x][z]=0.0;
	    }
	}
	ra0=atof(argv[argc-2])*M_PI/180.0;

	f4cbp=1.0/Dr/(double)N2*M_PI/(double)M;
{
#ifdef	OCT_SBS
	int	r0=N/2,
		r1=r0+N%2;
#else
	int	r0=abs(Ox),
		r1=hp.Nx;
#ifndef	OCT_SA
	int	r2=(Ox<0)?r0-1:r1,
		r3=(Ox<0)?r1:r0-1;
	double	r4=1.0/(double)(r1-r0+1);
#endif
#endif
	int	r00=(Ox<0)?Ox:0,
		r01=(Ox<0)?r0:0,
		r02=(Ox<0)?N:r1,
		r10=(Ox<0)?hp.Nx-1:N-1,
		r11=(Ox<0)?0:r1,
		r12=(Ox<0)?r0:N,
		z0=(Oy<0)?z2:z2+Oy,
		z1=(Oy<0)?z2-Oy:z2;

	for (m=0; m<M; m++) {
	    ReadHiPic(&hp,m);
	    for (z=0; z<L; z++) {
		p=P[z]; t=hp.T[z+z0]+r00;
		for (r=r01; r<r02; r++) p[r]=(-Log(t[r]));
	    }
	    ReadHiPic(&hp,m+M);
	    for (z=0; z<L; z++) {
		p=P[z]; t=hp.T[z+z1]+r10;
		for (r=r11; r<r12; r++) p[r]=(-Log(t[-r]));

		for (r=r0; r<r1; r++) p[r]=
#if	defined(OCT_SBS) || defined(OCT_SA)
		(p[r]-Log(t[-r]))/2.0;
#else
		((double)abs(r-r2)*p[r]+(double)abs(r-r3)*(-Log(t[-r])))*r4;
#endif
	    }
	    if (m>0)
		for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);

	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],Convolution,i);

	    sincos(ra0+M_PI*(double)m/(double)M,&S,&C); O=Or*(1.0-C+S);
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++)
		if (S4BP[y][x].j>=0) {
		    S4BP[y][x].j=r=(int)(R=C*(double)x-S*(double)y+O);
		    S4BP[y][x].r=R-(double)r;
		}

	    for (i=0; i<Threads; i++) TERM_MT(Thread[i],Convolution);
	    for (i=0; i<Threads; i++) INIT_MT(Thread[i],BackProjection,i);
	}
	for (i=0; i<Threads; i++) TERM_MT(Thread[i],BackProjection);
}
	free(*Q); free(Q); free(*P); free(P); free(*W); free(W);
	free(G); free(E); free(Thread);

	f=(FOM **)S4BP;
	for (z=0; z<L; z++) {
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++) f[y][x]=F[y][x][z];

	    (void)sprintf(path,argv[argc-1],z+z2);
	    StoreImageFile_Float(path,N,N,f,SIF_F_desc);
	    (void)printf("%d\t%s\n",z+z2,SIF_F_desc);
	}
	free(**F); free(*F); free(F); free(*S4BP); free(S4BP);
}
	TermReadHiPic(&hp);

	return 0;
}
